import logging

from algoliasearch.exceptions import AlgoliaException

from ..settings import DEBUG


logger = logging.getLogger(__name__)


class AlgoliaIndexError(Exception):
    """Something went wrong with an Algolia Index."""


class BaseAlgoliaIndex:

    # Use to specify the fields that should be included in the index.
    fields = ()

    # Use to specify the geo-fields that should be used for location search.
    # The attribute should be a callable that returns a tuple.
    geo_field = None

    # Use to specify the field that should be used for filtering by tag.
    tags = None

    # Use to specify the index to target on Algolia.
    index_name = None

    # Use to specify the settings of the index.
    settings = None

    # Used to specify if the instance should be indexed.
    # The attribute should be either:
    # - a callable that returns a boolean.
    # - a BooleanField
    # - a boolean property or attribute
    should_index = None

    def __init__(self, client, settings):
        self.__client = client
        self._named_fields = {}
        self._translate_fields = {}

        if self.settings is None:  # Only set settings if the actual index class does not define some
            self.settings = {}

        if isinstance(self.fields, str):
            self.fields = (self.fields,)
        elif isinstance(self.fields, (list, tuple, set)):
            self.fields = tuple(self.fields)
        else:
            raise AlgoliaIndexError('Fields must be a str, list, tuple or set')

    def _init_index(self, client, settings):

        tmp_index_name = '{index_name}_tmp'.format(index_name=self.index_name)

        if 'INDEX_PREFIX' in settings:
            self.index_name = settings['INDEX_PREFIX'] + '_' + self.index_name
            tmp_index_name = '{index_prefix}_{tmp_index_name}'.format(
                tmp_index_name=tmp_index_name,
                index_prefix=settings['INDEX_PREFIX']
            )
        if 'INDEX_SUFFIX' in settings:
            self.index_name += '_' + settings['INDEX_SUFFIX']
            tmp_index_name = '{tmp_index_name}_{index_suffix}'.format(
                tmp_index_name=tmp_index_name,
                index_suffix=settings['INDEX_SUFFIX']
            )

        self.__index = client.init_index(self.index_name)
        self.__tmp_index = client.init_index(tmp_index_name)

    @staticmethod
    def _validate_geolocation(geolocation):
        """
        Make sure we have the proper geolocation format.
        """
        if set(geolocation) != {'lat', 'lng'}:
            raise AlgoliaIndexError(
                'Invalid geolocation format, requires "lat" and "lng" keys only got {}'.format(
                    geolocation
                )
            )

    def get_raw_record(self, instance, update_fields=None):
        """
        Gets the raw record.

        If `update_fields` is set, the raw record will be build with only
        the objectID and the given fields. Also, `_geoloc` and `_tags` will
        not be included.
        """
        tmp = {'objectID': self.objectID(instance)}

        if update_fields:
            if isinstance(update_fields, str):
                update_fields = (update_fields,)

            for elt in update_fields:
                key = self._translate_fields.get(elt, None)
                if key:
                    tmp[key] = self._named_fields[key](instance)
        else:
            for key, value in self._named_fields.items():
                tmp[key] = value(instance)

            if self.geo_field:
                loc = self.geo_field(instance)

                if isinstance(loc, tuple):
                    tmp['_geoloc'] = {'lat': loc[0], 'lng': loc[1]}
                elif isinstance(loc, dict):
                    self._validate_geolocation(loc)
                    tmp['_geoloc'] = loc
                elif isinstance(loc, list):
                    [self._validate_geolocation(geo) for geo in loc]
                    tmp['_geoloc'] = loc

            if self.tags:
                if callable(self.tags):
                    tmp['_tags'] = self.tags(instance)
                if not isinstance(tmp['_tags'], list):
                    tmp['_tags'] = list(tmp['_tags'])

        return tmp

    def _has_should_index(self):
        """Return True if this AlgoliaIndex has a should_index method or attribute"""
        return self.should_index is not None

    def _should_index(self, instance):
        """Return True if the object should be indexed (including when self.should_index is not set)."""
        if self._has_should_index():
            return self._should_really_index(instance)
        else:
            return True

    def _should_really_index(self, instance):
        raise NotImplementedError

    def save_record(self, instance, update_fields=None, **kwargs):
        """Saves the record.

        If `update_fields` is set, this method will use partial_update_object()
        and will update only the given fields (never `_geoloc` and `_tags`).

        For more information about partial_update_object:
        https://github.com/algolia/algoliasearch-client-python#update-an-existing-object-in-the-index
        """
        if not self._should_index(instance):
            # Should not index, but since we don't now the state of the
            # instance, we need to send a DELETE request to ensure that if
            # the instance was previously indexed, it will be removed.
            self.delete_record(instance)
            return

        try:
            if update_fields:
                obj = self.get_raw_record(instance,
                                          update_fields=update_fields)
                result = self.__index.partial_update_object(obj)
            else:
                obj = self.get_raw_record(instance)
                result = self.__index.save_object(obj)
            logger.info('SAVE %s FROM %s', obj['objectID'], instance._meta.label)
            return result
        except AlgoliaException as e:
            if DEBUG:
                raise e
            else:
                logger.warning('%s FROM %s NOT SAVED: %s', obj['objectID'],
                               instance._meta.label, e)

    def delete_record(self, instance):
        """Deletes the record."""
        objectID = self.objectID(instance)
        try:
            self.__index.delete_object(objectID)
            logger.info('DELETE %s FROM %s', objectID, instance._meta.label)
        except AlgoliaException as e:
            if DEBUG:
                raise e
            else:
                logger.warning('%s FROM %s NOT DELETED: %s', objectID,
                               instance._meta.label, e)

    def update_records(self, qs, batch_size=1000, **kwargs):
        """
        Updates multiple records.

        This method is optimized for speed. It takes a QuerySet and the same
        arguments as QuerySet.update(). Optionnaly, you can specify the size
        of the batch send to Algolia with batch_size (default to 1000).

        >>> from algoliasearch_django import update_records
        >>> qs = MyModel.objects.filter(myField=False)
        >>> update_records(MyModel, qs, myField=True)
        >>> qs.update(myField=True)
        """
        tmp = {}
        for key, value in kwargs.items():
            name = self._translate_fields.get(key, None)
            if name:
                tmp[name] = value

        batch = []
        objectsIDs = qs.only(self.custom_objectID).values_list(
            self.custom_objectID, flat=True)
        for elt in objectsIDs:
            tmp['objectID'] = elt
            batch.append(dict(tmp))

            if len(batch) >= batch_size:
                self.__index.partial_update_objects(batch)
                batch = []

        if len(batch) > 0:
            self.__index.partial_update_objects(batch)

    def raw_search(self, query='', request_options=None):
        """Performs a search query and returns the parsed JSON."""
        if request_options is None:
            request_options = {}

        try:
            return self.__index.search(query, request_options=request_options)
        except AlgoliaException as e:
            if DEBUG:
                raise e
            else:
                logger.warning('ERROR DURING SEARCH ON %s: %s', self.index_name, e)

    def get_settings(self):
        """Returns the settings of the index."""
        try:
            logger.info('GET SETTINGS ON %s', self.index_name)
            return self.__index.get_settings()
        except AlgoliaException as e:
            if DEBUG:
                raise e
            else:
                logger.warning('ERROR DURING GET_SETTINGS ON %s: %s',
                               self.index_name, e)

    def set_settings(self):
        """Applies the settings to the index."""
        if not self.settings:
            return

        try:
            self.__index.set_settings(self.settings)
            logger.info('APPLY SETTINGS ON %s', self.index_name)
        except AlgoliaException as e:
            if DEBUG:
                raise e
            else:
                logger.warning('SETTINGS NOT APPLIED ON %s: %s',
                               self.index_name, e)

    def clear_objects(self):
        """Clears the index."""
        try:
            self.__index.clear_objects()
            logger.info('CLEAR INDEX %s', self.index_name)
        except AlgoliaException as e:
            if DEBUG:
                raise e
            else:
                logger.warning('%s NOT CLEARED: %s', self.index_name, e)

    def wait_task(self, task_id):
        try:
            self.__index.wait_task(task_id)
            logger.info('WAIT TASK %s', self.index_name)
        except AlgoliaException as e:
            if DEBUG:
                raise e
            else:
                logger.warning('%s NOT WAIT: %s', self.index_name, e)

    def reindex_all(self, batch_size=1000):
        """
        Reindex all the records.

        By default, this method use Model.objects.all() but you can implement
        a method `get_queryset` in your subclass. This can be used to optimize
        the performance (for example with select_related or prefetch_related).
        """
        should_keep_synonyms = False
        should_keep_rules = False
        try:
            if not self.settings:
                self.settings = self.get_settings()
                logger.debug('Got settings for index %s: %s', self.index_name, self.settings)
            else:
                logger.debug("index %s already has settings: %s", self.index_name, self.settings)
        except AlgoliaException as e:
            if any("Index does not exist" in arg for arg in e.args):
                pass  # Expected, let's clear and recreate from scratch
            else:
                raise e  # Unexpected error while getting settings
        try:
            if self.settings:
                replicas = self.settings.get('replicas', None)
                slaves = self.settings.get('slaves', None)

                should_keep_replicas = replicas is not None
                should_keep_slaves = slaves is not None

                if should_keep_replicas:
                    self.settings['replicas'] = []
                    logger.debug("REMOVE REPLICAS FROM SETTINGS")
                if should_keep_slaves:
                    self.settings['slaves'] = []
                    logger.debug("REMOVE SLAVES FROM SETTINGS")

                response = self.__tmp_index.set_settings(self.settings.copy())
                response.wait()
                logger.debug('APPLY SETTINGS ON %s_tmp', self.index_name)
            rules = []
            synonyms = []
            for r in self.__index.browse_rules():
                rules.append(r)
            for s in self.__index.browse_synonyms():
                synonyms.append(s)
            if len(rules):
                logger.debug('Got rules for index %s: %s', self.index_name, rules)
                should_keep_rules = True
            if len(synonyms):
                logger.debug('Got synonyms for index %s: %s', self.index_name, rules)
                should_keep_synonyms = True

            self.__tmp_index.clear_objects()
            logger.debug('CLEAR INDEX %s_tmp', self.index_name)

            counts = 0
            batch = []

            for instance in self.get_queryset():
                if not self._should_index(instance):
                    continue  # should not index

                batch.append(self.get_raw_record(instance))
                if len(batch) >= batch_size:
                    self.__tmp_index.save_objects(batch)
                    logger.info('SAVE %d OBJECTS TO %s_tmp', len(batch),
                                self.index_name)
                    batch = []
                counts += 1
            if len(batch) > 0:
                self.__tmp_index.save_objects(batch)
                logger.info('SAVE %d OBJECTS TO %s_tmp', len(batch),
                            self.index_name)

            self.__client.move_index(self.__tmp_index.name,
                                     self.__index.name)
            logger.info('MOVE INDEX %s_tmp TO %s', self.index_name,
                        self.index_name)

            if self.settings:
                if should_keep_replicas:
                    self.settings['replicas'] = replicas
                    logger.debug("RESTORE REPLICAS")
                if should_keep_slaves:
                    self.settings['slaves'] = slaves
                    logger.debug("RESTORE SLAVES")
                if should_keep_replicas or should_keep_slaves:
                    self.__index.set_settings(self.settings)
                if should_keep_rules:
                    response = self.__index.save_rules(
                            rules,
                            {"forwardToReplicas": True},
                        )
                    response.wait()
                    logger.info("Saved rules for index %s with response: {}".format(response), self.index_name)
                if should_keep_synonyms:
                    response = self.__index.save_synonyms(
                            synonyms,
                            {"forwardToReplicas": True},
                        )
                    response.wait()
                    logger.info("Saved synonyms for index %s with response: {}".format(response), self.index_name)
            return counts
        except AlgoliaException as e:
            if DEBUG:
                raise e
            else:
                logger.warning('ERROR DURING REINDEXING %s: %s', self.index_name,
                               e)



from django.core.management.base import BaseCommand

from algoliasearch_django import get_registered_adapters
from algoliasearch_django import reindex_all


class Command(BaseCommand):
    help = 'Reindex all models to Algolia'

    def add_arguments(self, parser):
        parser.add_argument('--batchsize', nargs='?', default=1000, type=int)
        parser.add_argument('--index', nargs='+', type=str)

    def handle(self, *args, **options):
        """Run the management command."""
        batch_size = options.get('batchsize', None)
        if not batch_size:
            # py34-django18: batchsize is set to None if the user don't set
            # the value, instead of not be present in the dict
            batch_size = 1000

        self.stdout.write('The following indexes were reindexed:')

        for adapter in get_registered_adapters():
            if options.get('index', None) is not None:
                if (adapter.index_name not in options['index']):
                    continue

            counts = adapter.reindex_all(batch_size=batch_size)
            self.stdout.write('\t* {} --> {}'.format(adapter.index_name, counts))



"""
AlgoliaSearch integration for Django.
http://www.algolia.com
"""

from django.utils.module_loading import autodiscover_modules

from . import models
from . import registration
from . import settings
from . import version

__version__ = version.VERSION
ALGOLIA_SETTINGS = settings.SETTINGS

AlgoliaIndex = models.AlgoliaIndex
Aggregator = models.Aggregator
AlgoliaEngine = registration.AlgoliaEngine
algolia_engine = registration.algolia_engine


# Algolia Engine functions

register = algolia_engine.register
register_aggregator = algolia_engine.register_aggregator
unregister = algolia_engine.unregister
get_registered_model = algolia_engine.get_registered_models

get_registered_adapters = algolia_engine.get_registered_adapters
get_adapter = algolia_engine.get_adapter
get_adapter_from_instance = algolia_engine.get_adapter_from_instance

save_record = algolia_engine.save_record
delete_record = algolia_engine.delete_record
update_records = algolia_engine.update_records
raw_search = algolia_engine.raw_search
clear_objects = algolia_engine.clear_objects
reindex_all = algolia_engine.reindex_all

# Default log handler
import logging


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def autodiscover():
    autodiscover_modules('index')


logging.getLogger(__name__.split('.')[0]).addHandler(NullHandler())

default_app_config = 'algoliasearch_django.apps.AlgoliaConfig'



from __future__ import unicode_literals

import inspect
from functools import partial
from itertools import chain
import logging
import sys

from django.db.models.query_utils import DeferredAttribute

from .base import BaseAlgoliaIndex, AlgoliaIndexError


logger = logging.getLogger(__name__)


def _getattr(obj, name):
    return getattr(obj, name)


def check_and_get_attr(model, name):
    try:
        attr = getattr(model, name)
        if callable(attr):
            return attr
        else:
            return get_model_attr(name)
    except AttributeError:
        raise AlgoliaIndexError(
            '{} is not an attribute of {}'.format(name, model))


def get_model_attr(name):
    return partial(_getattr, name=name)


class AlgoliaIndex(BaseAlgoliaIndex):
    """An index in the Algolia backend."""

    # Use to specify a custom field that will be used for the objectID.
    # This field should be unique.
    custom_objectID = 'pk'

    # Name of the attribute to check on instances if should_index is not a callable
    _should_index_is_method = False

    def __init__(self, model, client, settings):
        """Initializes the index."""
        if not self.index_name:
            self.index_name = model.__name__

        self._init_index(client, settings)

        super().__init__(client, settings)

        self.model = model

        all_model_fields = [f.name for f in model._meta.get_fields() if not f.is_relation]

        # Check fields
        for field in self.fields:
            if isinstance(field, str):
                attr = field
                name = field
            elif isinstance(field, (list, tuple)) and len(field) == 2:
                attr = field[0]
                name = field[1]
            else:
                raise AlgoliaIndexError(
                    'Invalid fields syntax: {} (type: {})'.format(field, type(field)))

            self._translate_fields[attr] = name
            if attr in all_model_fields:
                self._named_fields[name] = get_model_attr(attr)
            else:
                self._named_fields[name] = check_and_get_attr(model, attr)

        # If no fields are specified, index all the fields of the model
        if not self.fields:
            self.fields = set(all_model_fields)
            for elt in ('pk', 'id', 'objectID'):
                try:
                    self.fields.remove(elt)
                except KeyError:
                    continue
            self._translate_fields = dict(zip(self.fields, self.fields))
            self._named_fields = dict(zip(self.fields, map(get_model_attr,
                                                            self.fields)))

        # Check custom_objectID
        if self.custom_objectID in chain(['pk'], all_model_fields) or hasattr(model, self.custom_objectID):
            self.objectID = get_model_attr(self.custom_objectID)
        else:
            raise AlgoliaIndexError('{} is not a model field of {}'.format(
                self.custom_objectID, model))

        # Check tags
        if self.tags:
            if self.tags in all_model_fields:
                self.tags = get_model_attr(self.tags)
            else:
                self.tags = check_and_get_attr(model, self.tags)

        # Check geo_field
        if self.geo_field:
            self.geo_field = check_and_get_attr(model, self.geo_field)

        # Check should_index + get the callable or attribute/field name
        if self.should_index:
            if hasattr(model, self.should_index):
                attr = getattr(model, self.should_index)
                if type(attr) is not bool:  # if attr is a bool, we keep attr=name to getattr on instance
                    self.should_index = attr
                if callable(self.should_index):
                    self._should_index_is_method = True
            else:
                try:
                    model._meta.get_field_by_name(self.should_index)
                except:
                    raise AlgoliaIndexError('{} is not an attribute nor a field of {}.'.format(
                        self.should_index, model))

    def get_queryset(self):
        return self.model.objects.all()

    def get_raw_record(self, instance, update_fields=None):
        tmp = super().get_raw_record(instance, update_fields)
        logger.debug('BUILD %s FROM %s', tmp['objectID'], self.model)
        return tmp

    def _should_really_index(self, instance):
        """Return True if according to should_index the object should be indexed."""
        if self._should_index_is_method:
            is_method = inspect.ismethod(self.should_index)
            try:
                count_args = len(inspect.signature(self.should_index).parameters)
            except AttributeError:
                # noinspection PyDeprecation
                count_args = len(inspect.getargspec(self.should_index).args)

            if is_method or count_args is 1:
                # bound method, call with instance
                return self.should_index(instance)
            else:
                # unbound method, simply call without arguments
                return self.should_index()
        else:
            # property/attribute/Field, evaluate as bool
            attr_type = type(self.should_index)
            if attr_type is DeferredAttribute:
                attr_value = self.should_index.__get__(instance, None)
            elif attr_type is str:
                attr_value = getattr(instance, self.should_index)
            elif attr_type is property:
                attr_value = self.should_index.__get__(instance)
            else:
                raise AlgoliaIndexError('{} should be a boolean attribute or a method that returns a boolean.'.format(
                    self.should_index))
            if type(attr_value) is not bool:
                raise AlgoliaIndexError("%s's should_index (%s) should be a boolean" % (
                    instance.__class__.__name__, self.should_index))
            return attr_value



import sys

from functools import partial
from itertools import chain

from .base import BaseAlgoliaIndex, AlgoliaIndexError


def _getattr(obj, name):
    attr = getattr(obj, name, "")
    if callable(attr):
        return attr()
    else:
        return attr


def get_model_attr(name):
    return partial(_getattr, name=name)


class Aggregator(BaseAlgoliaIndex):

    custom_objectID = None

    def __init__(self, models, client, settings):
        if not self.index_name:
            self.index_name = type(self).__name__

        self._init_index(client, settings)

        super().__init__(client, settings)

        self.models = models
        all_model_fields = chain.from_iterable(
                [f.name for f in model._meta.get_fields() if not f.is_relation]
                for model in self.models)

        # Check fields
        for field in self.fields:
            if isinstance(field, str):
                attr = field
                name = field
            elif isinstance(field, (list, tuple)) and len(field) == 2:
                attr = field[0]
                name = field[1]
            else:
                raise AlgoliaIndexError(
                    'Invalid fields syntax: {} (type: {})'.format(field, type(field)))

            self._translate_fields[attr] = name
            self._named_fields[name] = get_model_attr(attr)

        # If no fields are specified, index all the fields of the model
        if not self.fields:
            self.fields = set(all_model_fields)
            for elt in ('pk', 'id', 'objectID'):
                try:
                    self.fields.remove(elt)
                except KeyError:
                    continue
            self._translate_fields = dict(zip(self.fields, self.fields))
            self._named_fields = dict(zip(self.fields, map(get_model_attr,
                                                            self.fields)))

        # Keep track of object model for easy faceting
        self._named_fields['_objectModel'] = self._model_name

        # Check custom_objectID
        if self.custom_objectID:
            self.custom_objectID = get_model_attr(self.custom_objectID)

        # Check tags
        if self.tags:
            self.tags = get_model_attr(self.tags)

        # Check geo_field
        if self.geo_field:
            self.geo_field = get_model_attr(self.geo_field)

        # Check should_index + get the callable or attribute/field name
        if self.should_index:
            self.should_index = get_model_attr(self.should_index)

    def _should_really_index(self, instance):
        """Return True if according to should_index the object should be indexed."""
        should_index = self.should_index(instance)
        if type(should_index) is not bool:
            raise AlgoliaIndexError("%s's should_index (%s) should be a boolean" % (
                instance.__class__.__name__, should_index))
        return should_index

    @staticmethod
    def _model_name(instance):
        return instance._meta.label

    def objectID(self, instance):
        if self.custom_objectID:
            return self.custom_objectID(instance)
        else:
            return "{}:{}".format(self._model_name(instance), instance.pk)

    def get_queryset(self):
        # This is not really a queryset, but behaves enough like one for reindex_all.
        return chain.from_iterable(model.objects.all() for model in self.models)




import logging

from django.db.models.signals import post_save
from django.db.models.signals import pre_delete
from algoliasearch.search_client import SearchClient, SearchConfig

from .models import AlgoliaIndex, Aggregator
from .settings import SETTINGS
from .version import VERSION
from algoliasearch.version import VERSION as CLIENT_VERSION
from platform import python_version
from django import get_version as django_version

logger = logging.getLogger(__name__)


class AlgoliaEngineError(Exception):
    """Something went wrong with Algolia Engine."""


class RegistrationError(AlgoliaEngineError):
    """Something went wrong when registering a model."""


class AlgoliaEngine:
    def __init__(self, settings=SETTINGS):
        """Initializes the Algolia engine."""

        try:
            app_id = settings['APPLICATION_ID']
            api_key = settings['API_KEY']
        except KeyError:
            raise AlgoliaEngineError(
                'APPLICATION_ID and API_KEY must be defined.')

        self.__auto_indexing = settings.get('AUTO_INDEXING', True)
        self.__settings = settings

        self.__registered_models = {}
        self.__registered_adapters = []

        config = SearchConfig(app_id, api_key)
        config.headers['User-Agent'] = 'Algolia for Python (%s); Python (%s); Algolia for Django (%s); Django (%s)' \
                                       % (CLIENT_VERSION, python_version(), VERSION, django_version)
        self.client = SearchClient.create_with_config(config)

    def is_registered(self, model):
        """Checks whether the given models is registered with Algolia engine"""
        return model in self.__registered_models

    def register(self, model, index_cls=AlgoliaIndex, auto_indexing=None):
        """
        Registers the given model with Algolia engine.

        If the given model is already registered with Algolia engine, a
        RegistrationError will be raised.
        """
        # Check for existing registration.
        if self.is_registered(model):
            raise RegistrationError(
                '{} is already registered with Algolia engine'.format(model))

        # Perform the registration.
        if not issubclass(index_cls, AlgoliaIndex):
            raise RegistrationError(
                '{} should be a subclass of AlgoliaIndex'.format(index_cls))
        index_obj = index_cls(model, self.client, self.__settings)
        self.__registered_models[model] = index_obj
        self.__registered_adapters.append(index_obj)

        if (isinstance(auto_indexing, bool) and
                auto_indexing) or self.__auto_indexing:
            # Connect to the signalling framework.
            post_save.connect(self.__post_save_receiver, model)
            pre_delete.connect(self.__pre_delete_receiver, model)
            logger.info('REGISTER %s', model)

    def register_aggregator(self, models, index_cls=Aggregator, auto_indexing=None):
        for model in models:
            # Check for existing registration.
            if self.is_registered(model):
                raise RegistrationError(
                    '{} is already registered with Algolia engine'.format(model))

        # Perform the registration.
        if not issubclass(index_cls, Aggregator):
            raise RegistrationError(
                '{} should be a subclass of Aggregator'.format(index_cls))

        index_obj = index_cls(models, self.client, self.__settings)
        self.__registered_adapters.append(index_obj)

        for model in models:
            self.__registered_models[model] = index_obj
            if (isinstance(auto_indexing, bool) and
                    auto_indexing) or self.__auto_indexing:
                # Connect to the signalling framework.
                post_save.connect(self.__post_save_receiver, model)
                pre_delete.connect(self.__pre_delete_receiver, model)
                logger.info('REGISTER %s', model)

    def unregister(self, model):
        """
        Unregisters the given model with Algolia engine.

        If the given model is not registered with Algolia engine, a
        RegistrationError will be raised.
        """
        if not self.is_registered(model):
            raise RegistrationError(
                '{} is not registered with Algolia engine'.format(model))
        # Perform the unregistration.
        del self.__registered_models[model]

        # Disconnect from the signalling framework.
        post_save.disconnect(self.__post_save_receiver, model)
        pre_delete.disconnect(self.__pre_delete_receiver, model)
        logger.info('UNREGISTER %s', model)

    def get_registered_models(self):
        """
        Returns a list of models that have been registered with Algolia
        engine.
        """
        return list(self.__registered_models.keys())

    def get_registered_adapters(self):
        """
        Returns a list of adapters that have been registered with Algolia
        engine.
        """
        return self.__registered_adapters

    def get_adapter(self, model):
        """Returns the adapter associated with the given model."""
        if not self.is_registered(model):
            raise RegistrationError(
                '{} is not registered with Algolia engine'.format(model))

        return self.__registered_models[model]

    def get_adapter_from_instance(self, instance):
        """Returns the adapter associated with the given instance."""
        model = instance.__class__
        return self.get_adapter(model)

    # Proxies methods.

    def save_record(self, instance, **kwargs):
        """Saves the record.

        If `update_fields` is set, this method will use partial_update_object()
        and will update only the given fields (never `_geoloc` and `_tags`).

        For more information about partial_update_object:
        https://github.com/algolia/algoliasearch-client-python#update-an-existing-object-in-the-index
        """
        adapter = self.get_adapter_from_instance(instance)
        adapter.save_record(instance, **kwargs)

    def delete_record(self, instance):
        """Deletes the record."""
        adapter = self.get_adapter_from_instance(instance)
        adapter.delete_record(instance)

    def update_records(self, model, qs, batch_size=1000, **kwargs):
        """
        Updates multiple records.

        This method is optimized for speed. It takes a QuerySet and the same
        arguments as QuerySet.update(). Optionally, you can specify the size
        of the batch send to Algolia with batch_size (default to 1000).

        >>> from algoliasearch_django import update_records
        >>> qs = MyModel.objects.filter(myField=False)
        >>> update_records(MyModel, qs, myField=True)
        >>> qs.update(myField=True)
        """
        adapter = self.get_adapter(model)
        adapter.update_records(qs, batch_size=batch_size, **kwargs)

    def raw_search(self, model, query='', request_options=None):
        """Performs a search query and returns the parsed JSON."""
        if request_options is None:
            request_options = {}

        adapter = self.get_adapter(model)
        return adapter.raw_search(query, request_options)

    def clear_objects(self, model):
        """Clears the index."""
        adapter = self.get_adapter(model)
        adapter.clear_objects()

    def reindex_all(self, model, batch_size=1000):
        """
        Reindex all the records.

        By default, this method use Model.objects.all() but you can implement
        a method `get_queryset` in your subclass. This can be used to optimize
        the performance (for example with select_related or prefetch_related).
        """
        adapter = self.get_adapter(model)
        return adapter.reindex_all(batch_size)

    def reset(self, settings=None):
        """Reinitializes the Algolia engine and its client.
        :param settings: settings to use instead of the default django.conf.settings.algolia
        """
        self.__init__(settings=settings if settings is not None else SETTINGS)

    # Signalling hooks.

    def __post_save_receiver(self, instance, **kwargs):
        """Signal handler for when a registered model has been saved."""
        logger.debug('RECEIVE post_save FOR %s', instance.__class__)
        self.save_record(instance, **kwargs)

    def __pre_delete_receiver(self, instance, **kwargs):
        """Signal handler for when a registered model has been deleted."""
        logger.debug('RECEIVE pre_delete FOR %s', instance.__class__)
        self.delete_record(instance)


# Algolia engine
algolia_engine = AlgoliaEngine()



from django.core.management.base import BaseCommand

from algoliasearch_django import get_registered_adapters
from algoliasearch_django import get_adapter


class Command(BaseCommand):
    help = 'Apply index settings.'

    def add_arguments(self, parser):
        parser.add_argument('--index', nargs='+', type=str)

    def handle(self, *args, **options):
        """Run the management command."""
        self.stdout.write('Apply settings to index:')
        for adapter in get_registered_adapters():
            if options.get('index', None) is not None:
                if (adapter.index_name not in options['index']):
                    continue

            adapter.set_settings()
            self.stdout.write('\t* {}'.format(adapter.index_name))



from contextlib import ContextDecorator
from functools import WRAPPER_ASSIGNMENTS, wraps

from django.db.models.signals import post_save, pre_delete

from . import algolia_engine


def available_attrs(fn):
    """
    Return the list of functools-wrappable attributes on a callable.
    This was required as a workaround for http://bugs.python.org/issue3445
    under Python 2.
    """
    return WRAPPER_ASSIGNMENTS


def register(model):
    """
    Register the given model class and wrapped AlgoliaIndex class with the Algolia engine:

    @register(Author)
    class AuthorIndex(AlgoliaIndex):
        pass

    """
    from algoliasearch_django import AlgoliaIndex, register

    def _algolia_engine_wrapper(index_class):
        if not issubclass(index_class, AlgoliaIndex):
            raise ValueError('Wrapped class must subclass AlgoliaIndex.')

        register(model, index_class)

        return index_class
    return _algolia_engine_wrapper


class disable_auto_indexing(ContextDecorator):
    """
    A context decorator to disable the auto-indexing behaviour of the AlgoliaIndex

    Can be used either as a context manager or a method decorator:
    >>> with disable_auto_indexing():
    >>>     my_object.save()

    >>> @disable_auto_indexing()
    >>> big_operation()
    """

    def __init__(self, model=None):
        if model is not None:
            self.models = [model]
        else:
            self.models = algolia_engine._AlgoliaEngine__registered_models

    def __enter__(self):
        for model in self.models:
            post_save.disconnect(
                algolia_engine._AlgoliaEngine__post_save_receiver,
                sender=model
            )
            pre_delete.disconnect(
                algolia_engine._AlgoliaEngine__pre_delete_receiver,
                sender=model
            )

    def __exit__(self, exc_type, exc_value, traceback):
        for model in self.models:
            post_save.connect(
                algolia_engine._AlgoliaEngine__post_save_receiver,
                sender=model
            )
            pre_delete.connect(
                algolia_engine._AlgoliaEngine__pre_delete_receiver,
                sender=model
            )



#!/usr/bin/env python

import os
import sys

from setuptools import setup
from setuptools import find_packages


# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

path_readme = os.path.join(os.path.dirname(__file__), 'README.md')
with open(path_readme) as readme:
    README = readme.read()

path_version = os.path.join(os.path.dirname(__file__),
                            'algoliasearch_django/version.py')

exec(open(path_version).read())

setup(
    name='brightnetwork-algoliasearch-django',
    version=VERSION,
    license='MIT License',
    packages=find_packages(exclude=['tests']),
    install_requires=['django>=2.2', 'algoliasearch>=2.0,<3.0'],
    description='Algolia Search integration for Django',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Algolia Team',
    author_email='support@algolia.com',
    url='https://github.com/brightnetwork/algoliasearch-django',
    keywords=['algolia', 'pyalgolia', 'search', 'backend', 'hosted', 'cloud',
              'full-text search', 'faceted search', 'django'],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
    ]
)



from django.apps import AppConfig


class AlgoliaConfig(AppConfig):
    """Simple AppConfig which does not do automatic discovery."""

    name = 'algoliasearch_django'

    def ready(self):
        super().ready()
        self.module.autodiscover()
